// Code generated by "stringer -type=endianess"; DO NOT EDIT.

package protocol

import "strconv"

const _endianess_name = "bigEndianlittleEndian"

var _endianess_index = [...]uint8{0, 9, 21}

func (i endianess) String() string {
	if i < 0 || i >= endianess(len(_endianess_index)-1) {
		return "endianess(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _endianess_name[_endianess_index[i]:_endianess_index[i+1]]
}
