// Code generated by "stringer -type=partKind"; DO NOT EDIT.

package protocol

import "strconv"

const _partKind_name = "pkNilpkCommandpkResultsetpkErrorpkStatementIDpkTransactionIDpkRowsAffectedpkResultsetIDpkTopologyInformationpkTableLocationpkReadLobRequestpkReadLobReplypkAbapIStreampkAbapOStreampkCommandInfopkWriteLobRequestpkClientContextpkWriteLobReplypkParameterspkAuthenticationpkSessionContextpkClientIDpkProfilepkStatementContextpkPartitionInformationpkOutputParameterspkConnectOptionspkCommitOptionspkFetchOptionspkFetchSizepkParameterMetadatapkResultMetadatapkFindLobRequestpkFindLobReplypkItabSHMpkItabChunkMetadatapkItabMetadatapkItabResultChunkpkClientInfopkStreamDatapkOStreamResultpkFDARequestMetadatapkFDAReplyMetadatapkBatchPreparepkBatchExecutepkTransactionFlagspkRowSlotImageParamMetadatapkRowSlotImageResultsetpkDBConnectInfopkLobFlagspkResultsetOptionspkXATransactionInfopkSessionVariablepkWorkLoadReplayContextpkSQLReplyOptions"

var _partKind_map = map[partKind]string{
	0:  _partKind_name[0:5],
	3:  _partKind_name[5:14],
	5:  _partKind_name[14:25],
	6:  _partKind_name[25:32],
	10: _partKind_name[32:45],
	11: _partKind_name[45:60],
	12: _partKind_name[60:74],
	13: _partKind_name[74:87],
	15: _partKind_name[87:108],
	16: _partKind_name[108:123],
	17: _partKind_name[123:139],
	18: _partKind_name[139:153],
	25: _partKind_name[153:166],
	26: _partKind_name[166:179],
	27: _partKind_name[179:192],
	28: _partKind_name[192:209],
	29: _partKind_name[209:224],
	30: _partKind_name[224:239],
	32: _partKind_name[239:251],
	33: _partKind_name[251:267],
	34: _partKind_name[267:283],
	35: _partKind_name[283:293],
	38: _partKind_name[293:302],
	39: _partKind_name[302:320],
	40: _partKind_name[320:342],
	41: _partKind_name[342:360],
	42: _partKind_name[360:376],
	43: _partKind_name[376:391],
	44: _partKind_name[391:405],
	45: _partKind_name[405:416],
	47: _partKind_name[416:435],
	48: _partKind_name[435:451],
	49: _partKind_name[451:467],
	50: _partKind_name[467:481],
	51: _partKind_name[481:490],
	53: _partKind_name[490:509],
	55: _partKind_name[509:523],
	56: _partKind_name[523:540],
	57: _partKind_name[540:552],
	58: _partKind_name[552:564],
	59: _partKind_name[564:579],
	60: _partKind_name[579:599],
	61: _partKind_name[599:617],
	62: _partKind_name[617:631],
	63: _partKind_name[631:645],
	64: _partKind_name[645:663],
	65: _partKind_name[663:690],
	66: _partKind_name[690:713],
	67: _partKind_name[713:728],
	68: _partKind_name[728:738],
	69: _partKind_name[738:756],
	70: _partKind_name[756:775],
	71: _partKind_name[775:792],
	72: _partKind_name[792:815],
	73: _partKind_name[815:832],
}

func (i partKind) String() string {
	if str, ok := _partKind_map[i]; ok {
		return str
	}
	return "partKind(" + strconv.FormatInt(int64(i), 10) + ")"
}
