// Code generated by "stringer -type=statementContextType"; DO NOT EDIT.

package protocol

import "strconv"

const _statementContextType_name = "scStatementSequenceInfoscServerExecutionTime"

var _statementContextType_index = [...]uint8{0, 23, 44}

func (i statementContextType) String() string {
	i -= 1
	if i < 0 || i >= statementContextType(len(_statementContextType_index)-1) {
		return "statementContextType(" + strconv.FormatInt(int64(i+1), 10) + ")"
	}
	return _statementContextType_name[_statementContextType_index[i]:_statementContextType_index[i+1]]
}
